@@ -10,16 +10,15 @@
 
 import bisect
 import concurrent
-import copy
 import glob
 import os
+import subprocess
 import sys
 import threading
 import time
-from collections import defaultdict, namedtuple
+from collections import OrderedDict, defaultdict, namedtuple
 from concurrent.futures import ThreadPoolExecutor
 from pathlib import Path
-from subprocess import Popen
 
 import flux
 import flux.importer
@@ -29,6 +28,72 @@ from flux.utils import tomli as tomllib
 from flux.utils.graphlib import TopologicalSorter
 
 
+def run_all_rc_scripts(runlevel):
+    """
+    Helper script for flux-modprobe(1) rc1 and rc3 scripts that replaces
+    the following shell code from rc1/rc3:
+    ```
+    core_dir=$(cd ${0%/*} && pwd -P)
+    all_dirs=$core_dir${FLUX_RC_EXTRA:+":$FLUX_RC_EXTRA"}
+    IFS=:
+    for rcdir in $all_dirs; do
+        for rcfile in $rcdir/rc{runlevel}.d/*; do
+        [ -e $rcfile ] || continue
+            echo running $rcfile
+            $rcfile || exit_rc=1
+        done
+    done
+    ```
+
+    Args:
+        runlevel (int): runlevel (1 or 3) in which function is running
+    Raises:
+        OSError: one or more rc scripts failed
+    """
+    success = True
+    core_dir = Path(conf_builtin_get("confdir")).resolve()
+    all_dirs = [core_dir]
+    rc_extra = os.environ.get("FLUX_RC_EXTRA")
+    if rc_extra:
+        all_dirs.extend(Path(d) for d in rc_extra.split(":") if d.strip())
+
+    for entry in all_dirs:
+        rcdir = entry / f"rc{runlevel}.d"
+        if not rcdir.exists() or not rcdir.is_dir():
+            continue
+        try:
+            # Get all files in rcX.d directory, sorted by name
+            rc_files = sorted(
+                [
+                    file
+                    for file in rcdir.iterdir()
+                    if file.is_file() and os.access(file, os.X_OK)
+                ]
+            )
+
+            # for rcfile in $rcdir/rc1.d/*; do
+            for rcfile in rc_files:
+                try:
+                    print(f"running {rcfile}")
+                    subprocess.run([str(rcfile)], check=True)
+                except subprocess.CalledProcessError as e:
+                    success = False
+                    print(
+                        f"{rcfile} failed with exit code {e.returncode}",
+                        file=sys.stderr,
+                    )
+                except (FileNotFoundError, PermissionError, OSError) as e:
+                    success = False
+                    print(f"Cannot execute {rcfile}: {e}", file=sys.stderr)
+
+        except (PermissionError, OSError) as e:
+            success = False
+            print(f"Cannot access directory {rcdir}: {e}", file=sys.stderr)
+
+    if not success:
+        raise OSError(f"one or more rc{runlevel}.d scripts failed")
+
+
 def default_flux_confdir():
     """
     Return the builtin Flux confdir
@@ -136,6 +201,27 @@ class Task:
         # convert self.ranks to rank conditional object:
         self.ranks = rank_conditional(self.ranks)
 
+    @staticmethod
+    def _check_spec(spec, getter):
+        """
+        Check if an identifier is set or unset based on an input ``spec``
+        and a ``getter`` function which should return the value of a ``spec``
+        (e.g. a config key or broker attribute) or None if the spec is
+        unset.
+
+        If ``spec`` begins with ``!``, then this function returns True only
+        if getter returns None for ``spec[1:]``. Otherwise, this function
+        returns False if the getter returns None for ``spec``.
+
+        Examples:
+          _check_spec("foo", getter) -> True if getter("foo") returns non-None
+          _check_spec("!foo", getter) -> True if getter("foo") returns None
+        """
+        inverted = spec.startswith("!")
+        attr = spec[1:] if inverted else spec
+        value = getter(attr)
+        return value is None if inverted else value is not None
+
     def enabled(self, context=None):
         """
         Return True if task is currently not disabled. A task may be
@@ -151,12 +237,10 @@ class Task:
         if self.disabled or not self.ranks.test(context.rank):
             return False
         for key in self.needs_config:
-            val = context.handle.conf_get(key)
-            if not val:
+            if not self._check_spec(key, context.handle.conf_get):
                 return False
         for attr in self.needs_attrs:
-            val = context.attr_get(attr)
-            if not val:
+            if not self._check_spec(attr, context.attr_get):
                 return False
         return True
 
@@ -227,6 +311,7 @@ class Module(Task):
         "needs-config",
         "priority",
         "disabled",
+        "exec",
     )
 
     def __init__(self, conf):
@@ -241,6 +326,7 @@ class Module(Task):
 
         self.module = conf.get("module", name)
         self.args = conf.get("args", [])
+        self.exec = conf.get("exec", False)
         self.run = self._load
 
         # Build kwargs to pass along to Task class
@@ -268,7 +354,7 @@ class Module(Task):
 
     def _load(self, context):
         args = context.getopts(self.name, default=self.args, also=self.provides)
-        payload = {"path": self.module, "args": args}
+        payload = {"path": self.module, "args": args, "exec": self.exec}
 
         if self.name != self.module:
             payload["name"] = self.name
@@ -291,7 +377,7 @@ class Module(Task):
             module_args = context.getopts(
                 self.name, default=self.args, also=self.provides
             )
-            print(f"load {self.name}", " ".join(module_args))
+            print(" ".join([f"load {self.name}", *module_args]))
         else:
             print(f"remove {self.name}")
 
@@ -418,8 +504,8 @@ def task(name, **kwargs):
         requires (optional, list): An optional list of task or module names
             this task requires. This is used to ensure required tasks are
             active when activating another task. It does not indicate that
-            this task will necessarily be run before the tasks it requires.
-            (See ``before`` for that feature)
+            this task will necessarily be run before or after the tasks it
+            requires. (See ``before`` or ``after`` for those features)
         needs (options, list): Disable this task if any task in ``needs`` is
             not active.
         provides (optional, list): An optional list of string service name
@@ -535,7 +621,14 @@ class Context:
 
     def bash(self, command):
         """Execute command under ``bash -c``"""
-        Popen(["bash", "-c", command]).wait()
+        process = subprocess.run(["bash", "-c", command])
+        if process.returncode != 0:
+            if process.returncode > 0:
+                raise RuntimeError(
+                    f"bash: exited with exit status {process.returncode}"
+                )
+            else:
+                raise RuntimeError(f"bash: died by signal {process.returncode}")
 
     def load_modules(self, modules):
         """Set a list of modules to load by name"""
@@ -570,9 +663,10 @@ class ModuleList:
         self.loaded_modules = []
         self.servicemap = {}
         for entry in resp["mods"]:
-            self.loaded_modules.append(entry["name"])
-            for name in entry["services"]:
-                self.servicemap[name] = entry["name"]
+            if entry["path"] != "builtin":
+                self.loaded_modules.append(entry["name"])
+                for name in entry["services"]:
+                    self.servicemap[name] = entry["name"]
 
     def __iter__(self):
         for name in self.loaded_modules:
@@ -587,7 +681,7 @@ class Modprobe:
     The modprobe main class. Intended for use by flux-modprobe(1).
     """
 
-    def __init__(self, confdir=None, timing=False, verbose=False, dry_run=False):
+    def __init__(self, timing=False, verbose=False, dry_run=False):
         self.exitcode = 0
         self.timing = None
         self.t0 = None
@@ -598,10 +692,10 @@ class Modprobe:
         self.handle = self.context.handle
         self.rank = self.handle.get_rank()
 
-        if confdir is None:
-            self.confdir = default_flux_confdir() / "modprobe"
-        else:
-            self.confdir = Path(confdir)
+        self.searchpath = {
+            "toml": self._get_searchpath(),
+            "py": self._get_searchpath(builtindir="libexecdir"),
+        }
 
         # Active tasks are those added via the @task decorator, and
         # which will be active by default when running "all" tasks:
@@ -629,6 +723,10 @@ class Modprobe:
             )
         )
 
+    def print(self, *args):
+        """Wrapper for context.print()"""
+        self.context.print(*args)
+
     def add_timing(self, name, starttime, end=None):
         if self.timing is None:
             return
@@ -669,10 +767,14 @@ class Modprobe:
         except ValueError:
             return False
 
-    def update_task(self, name, entry):
+    def update_module(self, name, entry, new_module=None):
         task = self.get_task(name)
-        for key, value in entry.items():
-            setattr(task, key, value)
+        if new_module is None:
+            if "name" not in entry:
+                entry["name"] = name
+            new_module = Module(entry)
+        for key in entry.keys():
+            setattr(task, key, getattr(new_module, key))
         self.taskdb.update(task)
 
     def add_modules(self, file):
@@ -683,35 +785,94 @@ class Modprobe:
             if name == "modules":
                 for table in entry:
                     try:
-                        self.add_task(Module(table))
+                        task = Module(table)
                     except ValueError as exc:
                         raise ValueError(
                             f"{file}: invalid modules entry: {exc}"
                         ) from None
+
+                    # Update tasks that already exist:
+                    if self.has_task(task.name):
+                        self.update_module(task.name, table, task)
+                    else:
+                        self.add_task(task)
             else:
                 # Allow <module>.key to update an existing configured module:
-                self.update_task(name, entry)
+                self.update_module(name, entry)
 
-    def _modprobe_path_expand(self, path=None, name="modprobe", ext="toml"):
-        if path is None:
-            etc = Path(self.confdir)
-            path = etc / f"{name}.{ext}"
-        else:
-            etc = Path(path).parent
-
-        files = [path]
-        dirs = [etc]
-        dirs.extend(
-            filter(
+    def _get_searchpath(self, builtindir="datadir"):
+        """
+        Return list of dirs in ``FLUX_MODPROBE_PATH`` if set, o/w returns the
+        default modprobe search path.
+        Args:
+            builtindir (str): base path for builtin/package path. Should
+                be either "datadir" or "libexecdir".
+        """
+        searchpath = []
+        if "FLUX_MODPROBE_PATH" in os.environ:
+            searchpath = filter(
                 lambda s: s and not s.isspace(),
-                os.environ.get("FLUX_MODPROBE_PATH", "").split(":"),
+                os.environ["FLUX_MODPROBE_PATH"].split(":"),
             )
-        )
-        for directory in dirs:
+        else:
+            pkgdir = conf_builtin_get(builtindir)
+            confdir = conf_builtin_get("confdir")
+            searchpath = [f"{pkgdir}/modprobe", f"{confdir}/modprobe"]
+
+        if "FLUX_MODPROBE_PATH_APPEND" in os.environ:
+            searchpath.extend(
+                filter(
+                    lambda s: s and not s.isspace(),
+                    os.environ["FLUX_MODPROBE_PATH_APPEND"].split(":"),
+                )
+            )
+
+        # return searchpath without duplicates
+        return list(OrderedDict.fromkeys(searchpath))
+
+    def _searchpath_expand(self, name="modprobe", ext="toml"):
+        """
+        Expand searchpath for extension ``ext`` based on configured paths.
+        """
+        files = []
+        for directory in self.searchpath[ext]:
+            self.print(f"checking {directory}/{name}.d/*.{ext}")
             if Path(directory).exists():
                 files.extend(sorted(glob.glob(f"{directory}/{name}.d/*.{ext}")))
         return files
 
+    def _get_toml_files(self):
+        """
+        Return all modprobe config toml files found in the following order
+         - Always read ``{fluxdatadir}/modprobe/modprobe.toml``
+         - for dir in self.searchpath: read ``{dir}/modprobe.d/*.toml``
+        """
+        files = []
+        builtin_toml_config = (
+            Path(conf_builtin_get("datadir")) / "modprobe" / "modprobe.toml"
+        )
+        self.print(f"checking {builtin_toml_config}")
+        if builtin_toml_config.exists():
+            files.append(str(builtin_toml_config))
+        files.extend(self._searchpath_expand())
+        return files
+
+    def _get_rc_files(self, name="rc1"):
+        """
+        Return all modprobe rc *.py files found in the following order
+         - Always read ``{fluxdatadir}/modprobe/{name}.py`` (e.g. ``rc1.py``)
+         - for dir in self.searchpath: read ``{dir}/{name}.d/*.py``
+        """
+        files = []
+        builtin_rc_file = (
+            Path(conf_builtin_get("libexecdir")) / "modprobe" / f"{name}.py"
+        )
+        self.print(f"checking {builtin_rc_file}")
+        if builtin_rc_file.exists():
+            files.append(str(builtin_rc_file))
+        files.extend(self._searchpath_expand(name=name, ext="py"))
+        return files
+
     def _update_modules_from_config(self):
         """Update modules using broker config
         Process a [modules] table in config support the following keys:
@@ -727,18 +888,14 @@ class Modprobe:
                 for service, name in entry.items():
                     self.taskdb.set_alternative(service, name)
             else:
-                self.update_task(key, entry)
+                self.update_module(key, entry)
 
-    def configure_modules(self, path=None):
-        """Load all configured modules in searchpath
-
-        Loads all modules TOML files in the following order:
-            1. From ``path`` if provided.
-               default path=``{sysconfdir}/modprobe/modprobe.toml``
-            2. From ``dirname(path)/modprobe.d/*.toml``
-            3. for dir in FLUX_MODPROBE_PATH: ``{dir}/modprobe.d/*.toml``
+    def configure_modules(self):
         """
-        for file in self._modprobe_path_expand(path):
+        Load module configuration from TOML config.
+        """
+        for file in self._get_toml_files():
+            self.print(f"loading {file}")
             self.add_modules(file)
 
         self._update_modules_from_config()
@@ -843,10 +1000,23 @@ class Modprobe:
                     if successor in deps:
                         deps[successor].append(task.name)
 
-    def get_deps(self, tasks, reverse=False):
+    def get_deps(self, tasks):
         """Return dependencies for tasks as dict of names to predecessor list"""
         t0 = self.timestamp
+        if not isinstance(tasks, set):
+            tasks = set(tasks)
         deps = {}
+
+        # Ensure tasks set contains all provides and the actual task name
+        # (since presence in the set determines if a task is included in
+        #  the predecessor list below)
+        provides = set()
+        for task in tasks:
+            task = self.get_task(task)
+            provides.add(task.name)
+            provides.update(task.provides)
+        tasks.update(provides)
+
         for name in tasks:
             task = self.get_task(name)
             if "*" in task.after:
@@ -861,6 +1031,22 @@ class Modprobe:
 
         return deps
 
+    def get_requires(self, tasks, reverse=False):
+        """Return dependencies for tasks as dicts of names to dependencies"""
+        deps = {}
+        for name in tasks:
+            task = self.get_task(name)
+            deps[task.name] = list(task.requires)
+        if reverse:
+            rdeps = {}
+            for dependent, reqs in deps.items():
+                for req in reqs:
+                    if req not in rdeps:
+                        rdeps[req] = set()
+                    rdeps[req].add(dependent)
+            return rdeps
+        return deps
+
     def run(self, deps):
         """Run all tasks in deps in precedence order"""
         t0 = self.timestamp
@@ -878,13 +1064,18 @@ class Modprobe:
         executor = ThreadPoolExecutor(max_workers=max_workers)
         futures = {}
         started = {}
+        aborted = False
 
         while sorter.is_active():
             for task in [self.get_task(x) for x in sorter.get_ready()]:
                 if task.name not in started:
-                    future = executor.submit(task.runtask, self.context)
-                    started[task.name] = task
-                    futures[future] = task
+                    if not aborted:
+                        future = executor.submit(task.runtask, self.context)
+                        started[task.name] = task
+                        futures[future] = task
+                    else:
+                        # Do not run task, mark it done
+                        sorter.done(task.name)
 
             done, not_done = concurrent.futures.wait(
                 futures.keys(), return_when=concurrent.futures.FIRST_COMPLETED
@@ -897,10 +1088,15 @@ class Modprobe:
                 except Exception as exc:
                     print(f"{task.name}: {exc}", file=sys.stderr)
                     self.exitcode = 1
+                    aborted = True
+                    # cancel any non-running but scheduled futures immediately
+                    for x in not_done:
+                        x.cancel()
                 sorter.done(task.name)
                 del futures[future]
 
         self.save_task_timing(started.values())
+        executor.shutdown(wait=True)
 
         return self.exitcode
 
@@ -918,12 +1114,13 @@ class Modprobe:
     def read_rcfile(self, name):
         # For absolute file path, just add tasks from single file:
         if name.endswith(".py"):
+            self.print(f"loading {name}")
             self._load_file(name)
             return
 
-        # O/w, add tasks from "named" file in confdir, plus all tasks
-        # in {confdir}/{name}.d/*.py and FLUX_MODPROBE_PAtH
-        for file in self._modprobe_path_expand(name=name, ext="py"):
+        # O/w, load all rc files in configured search path:
+        for file in self._get_rc_files(name):
+            self.print(f"loading {file}")
             self._load_file(file)
 
     def activate_modules(self, modules):
@@ -932,10 +1129,113 @@ class Modprobe:
             if not isinstance(task, Module):
                 raise ValueError(f"{module} is not a module")
             self._active_tasks.append(module)
+            # append any requires from this module
+            for other in task.requires:
+                self._active_tasks.append(other)
+
+    def _set_all_alternatives(self, modules):
+        # Set all modules as the current selected alternatives:
+        for module in modules:
+            task = self.get_task(module)
+            for service in task.provides:
+                self.set_alternative(service, task.name)
 
     def load(self, modules):
-        """Load modules and their dependencies"""
-        self.run(self.get_deps(self.solve(modules)))
+        """
+        Load modules and their dependencies (if not already loaded)
+
+        Args:
+            modules (list): List of modules to load.
+
+        Raises:
+            FileExistsError: Target modules (and all their dependencies)
+                are already loaded, so there is nothing to do.
+        """
+        mlist = ModuleList(self.handle)
+        needed_modules = [x for x in self.solve(modules) if x not in mlist]
+
+        # Ensure explicitly requested modules are the current alternatives
+        self._set_all_alternatives(needed_modules)
+
+        if needed_modules:
+            self.run(self.get_deps(needed_modules))
+        else:
+            raise FileExistsError(
+                "All modules and their dependencies are already loaded."
+            )
+
+    def _find_removable(self, dependencies, modules_to_remove):
+        """
+        Given a set of modules and dependency list of all modules, find modules
+        that can be removed because they no longer have any dependents.
+
+        Args:
+            dependencies (dict): Dictionary of modules to dependency list
+            modules_to_remove (list): List of modules to remove
+
+        Returns:
+            list: modules that can be safely removed (including original list)
+        """
+        dependents = {}
+        for dependent, reqs in dependencies.items():
+            for req in reqs:
+                if req not in dependents:
+                    dependents[req] = set()
+                dependents[req].add(dependent)
+
+        # Start with the items we're told to remove
+        removed_items = set(modules_to_remove)
+        newly_removable = []
+
+        # Keep track of items to check in this iteration
+        modules_to_check = set(modules_to_remove)
+
+        while modules_to_check:
+            next_modules_to_check = set()
+
+            for removed_item in modules_to_check:
+                # Find all dependencies of the removed item
+                # (items it depended on)
+                if removed_item in dependencies:
+                    for dependency in dependencies[removed_item]:
+                        # Skip if this dependency is already being removed
+                        if dependency in removed_items:
+                            continue
+
+                        # Check if this dependency still has other items
+                        # depending on it
+                        remaining_dependents = (
+                            dependents.get(dependency, set()) - removed_items
+                        )
+
+                        # If no remaining dependents, it can be removed
+                        if not remaining_dependents:
+                            removed_items.add(dependency)
+                            newly_removable.append(dependency)
+                            next_modules_to_check.add(dependency)
+
+            modules_to_check = next_modules_to_check
+
+        # Add newly removed items to list of items to remove
+        modules_to_remove.extend(newly_removable)
+
+        # Discard elements from dependents if they are being removed
+        for name in modules_to_remove:
+            for deps in dependents.values():
+                # discard both the name to remove and the real name
+                # of the task in case they are different
+                # (e.g. sched vs sched-simple):
+                deps.discard(name)
+                deps.discard(self.get_task(name).name)
+
+        # Raise an error if any removed modules still have dependents
+        for name in modules_to_remove:
+            if dependents.get(name):
+                raise ValueError(
+                    f"{name} still in use by " + ", ".join(dependents[name])
+                )
+
+        return modules_to_remove
 
     def _solve_modules_remove(self, modules=None):
         """Solve for a set of currently loaded modules to remove"""
@@ -951,44 +1251,25 @@ class Modprobe:
                 if not mlist.lookup(module):
                     raise ValueError(f"module {module} is not loaded")
 
-        # Calculate reverse deps for all loaded modules
-        deps = self.get_deps(all_modules)
-        rdeps = defaultdict(set)
+        modules = self._find_removable(self.get_requires(all_modules), modules)
 
+        # Compute reverse precedence graph of modules to remove so that
+        # they can be removed in reverse order of load:
+        deps = self.get_deps(modules)
+        rdeps = defaultdict(set)
         for name, deplist in deps.items():
             for mod in deplist:
-                rdeps[mlist.lookup(mod)].add(name)
-
-        # Determine if removal of modules will cause any other modules
-        # to become "unused", i.e. not needed by other modules
-        rdeps_copy = copy.deepcopy(rdeps)
-
-        def remove_dep(name):
-            for modname, dep_set in rdeps_copy.items():
-                if not dep_set:
-                    continue
-                dep_set.discard(name)
-                if not dep_set:
-                    remove_dep(modname)
-                    if modname in mlist:
-                        modules.append(modname)
-
-        for name in modules:
-            remove_dep(mlist.lookup(name))
-
-        # Raise an error if any module in modules has rdeps remaining
-        for name in modules:
-            if rdeps_copy[name]:
-                raise ValueError(
-                    f"{name} still in use by " + ", ".join(rdeps_copy[name])
-                )
+                mod = mlist.lookup(mod)
+                rdeps[mod].add(name)
 
+        # Convert module names to ModuleRemove tasks:
         tasks = set()
         for service in modules:
             name = mlist.lookup(service)
             if name is not None:
                 tasks.add(name)
 
+        # filter out tasks from rdeps that are not slated for removal
         deps = {}
         for name in tasks:
             deps[name] = {x for x in rdeps[name] if x in tasks}
@@ -1001,6 +1282,12 @@ class Modprobe:
             mlist = ModuleList(self.handle)
             modules = [x for x in mlist if self.has_task(x)]
 
+        # When removing modules, always set available alternatives to
+        # the specific modules being requested to remove. This prevents
+        # non-loaded but default alternatives from appearing in get_deps()
+        # later:
+        self._set_all_alternatives(modules)
+
         for module in modules:
             task = self.get_task(module)
             task.set_remove()
