@@ -810,6 +810,16 @@ class OutputFormat:
             fields.update([x for x, y in self.sort_keys])
         return list(fields)
 
+    @staticmethod
+    def _calculate_width(width):
+        if width == 0:
+            try:
+                width = os.get_terminal_size().columns
+            except OSError:
+                # No terminal found (or other error), set width to unlimited:
+                width = -1
+        return width
+
     # This should be a method, not a function since it is overridden by
     # inheriting classes
     # pylint: disable=no-self-use
@@ -845,10 +855,21 @@ class OutputFormat:
         fmt = "".join(format_list)
         return fmt
 
-    def header(self):
-        """Return a header row using the current format"""
+    def header(self, width=-1):
+        """
+        Return a header row using the current format
+        Args:
+            width (integer): If set > 0, truncate output after width columns.
+                if 0, then width is set to the current terminal width as
+                returned from :py:func:`os.get_terminal_size`. If < 0 (the
+                default), then no truncation occurs.
+        """
         formatter = self.HeaderFormatter()
-        return formatter.format(self.header_format(), **self.headings)
+        result = formatter.format(self.header_format(), **self.headings)
+        width = self._calculate_width(width)
+        if width > 0 and len(result) > width:
+            result = result[: width - 1] + "+"
+        return result
 
     def get_format_prepended(
         self,
@@ -948,7 +969,7 @@ class OutputFormat:
             raise KeyError(f"Invalid format field {exc} for {typestr}")
         return retval
 
-    def filter(self, items):
+    def filter(self, items, no_header=False):
         """
         Check for format fields that are prefixed with `?:` (e.g. "?:{name}"),
         and filter them out of the current format string if they result in an
@@ -960,6 +981,10 @@ class OutputFormat:
 
         (`?+:` requests both actions: filter out field if it is empty for all
         items, if not expand to maximum width)
+
+        Args:
+            items (iterable): list of items to consider for output
+            no_header (boolean): do not use header in calculations (default: False)
         """
 
         #  Build a list of all format strings that have one of the width
@@ -1000,12 +1025,18 @@ class OutputFormat:
 
                 #  Save the modified format, index, type, maximum width,
                 #  observed width, and broken-down spec in lst:
+                initialmaxwidth = spec.width or 0
+                if sentinels[end] in ("maxwidth", "both"):
+                    initialmaxwidth = max(
+                        initialmaxwidth,
+                        0 if no_header else len(self.headings.get(field, "")),
+                    )
                 lst.append(
                     dict(
                         fmt=fmt,
                         index=index,
                         type=sentinels[end],
-                        maxwidth=spec.width or 0,
+                        maxwidth=initialmaxwidth,
                         width=0,
                         spec=spec,
                     )
@@ -1107,7 +1138,7 @@ class OutputFormat:
             items.sort(key=attrgetter(key), reverse=reverse)
         return items
 
-    def print_items(self, items, no_header=False, pre=None, post=None):
+    def print_items(self, items, no_header=False, width=-1, pre=None, post=None):
         """
         Handle printing a list of items with the current format.
 
@@ -1125,11 +1156,15 @@ class OutputFormat:
         Args:
             items (iterable): list of items to format
             no_header (boolean): disable header row (default: False)
+            width (integer): If set > 0, truncate output after width
+              columns. If 0, then width is set to the current terminal
+              width as returned from :py:func:`os.get_terminal_size`.
+              if < 0 (the default), then no truncation occurs.
             pre (callable): Function to call before printing each item
             post (callable): Function to call after printing each item
         """
         #  Preprocess original format by processing with filter():
-        newfmt = self.filter(items)
+        newfmt = self.filter(items, no_header=no_header)
         #  Get the current class for creating a new formatter instance:
         cls = self.__class__
         #  Create new instance of the current class from filtered format:
@@ -1137,12 +1172,16 @@ class OutputFormat:
 
         items = self.sort_items(items)
 
+        width = self._calculate_width(width)
+
         if not no_header:
-            print(formatter.header())
+            print(formatter.header(width=width))
         for item in items:
             line = formatter.format(item)
             if not line or line.isspace():
                 continue
+            if width > 0 and len(line) > width:
+                line = line[: width - 1] + "+"
             if callable(pre):
                 pre(item)
             try:
@@ -1647,7 +1686,10 @@ class UtilConfig:
         try:
             return formats[format_name]["format"]
         except KeyError:
-            raise ValueError(f"--format: No such format {format_name}")
+            s = re.sub(r"\w+", r"{\g<0>}", format_name)
+            raise ValueError(
+                f"--format: No such format {format_name} (Did you mean '{s}'?)"
+            )
 
     def __getattr__(self, attr):
         return self.config[attr]
